#!/usr/bin/env python3
import unittest
import numpy as np
import brille as br_py
from brille import _brille as br_mod


class Lattice(unittest.TestCase):
    def setUp(self) -> None:
        self.cmo_vectors = np.array([[1.7797736800000001, 1.027552813333333, 6.219738366666666],
                                     [-1.7797736800000001, 1.027552813333333, 6.219738366666666],
                                     [0.0, -2.055105626666667, 6.219738366666666]])
        self.cmo_positions = np.array([[0.89401258, 0.89401259, 0.89401258],
                                       [0.10598742, 0.10598741, 0.10598742],
                                       [0.5, 0.5, 0.5],
                                       [0., 0.99999999, 0.]])
        self.cmo_types = [0, 0, 1, 2]
        self.cmo_rotations = np.array([[[-1, 0, 0], [0, -1, 0], [0, 0, -1]],
                                       [[0, 0, 1], [1, 0, 0], [0, 1, 0]],
                                       [[0, 0, -1], [0, -1, 0], [-1, 0, 0]]])
        self.cmo_translations = np.array([[0., 0., 0.] for _ in self.cmo_rotations])

        self.cmo_avg_positions = np.array([[0.8940125833333333, 0.8940125833333333, 0.8940125833333333],
                                           [0.1059874166666667, 0.1059874166666666, 0.1059874166666667],
                                           [0.5, 0.5, 0.5],
                                           [0., 1., 0.]])

    def test_a_init2(self):
        a = 1.0
        b = np.pi / 2
        v = 1.0
        l = br_mod.Lattice([a, a, a], [b, b, b])  # compute the volume when required
        self.assertEqual(l.a, a)
        self.assertEqual(l.b, a)
        self.assertEqual(l.c, a)
        self.assertEqual(l.alpha, b)
        self.assertEqual(l.beta, b)
        self.assertEqual(l.gamma, b)
        self.assertAlmostEqual(l.volume, v)  # since 0.9999999999999999 != 1.0

    def test_a_init3(self):
        a = (1, 1, 1)
        b = np.array([1, 1, 1]) * np.pi / 2
        v = 1.0
        l = br_mod.Lattice(a, b)
        self.assertEqual(l.a, a[0])
        self.assertEqual(l.b, a[1])
        self.assertEqual(l.c, a[2])
        self.assertEqual(l.alpha, b[0])
        self.assertEqual(l.beta, b[1])
        self.assertEqual(l.gamma, b[2])
        self.assertAlmostEqual(l.volume, v)  # since 0.9999999999999999 != 1.0

    def test_b_tensors(self):
        a = (6, 6, 9)
        b = np.array([90, 90, 120]) * np.pi / 180
        l = br_mod.Lattice(a, b)
        cmt = np.array([[a[0] ** 2, a[0] * a[1] * np.cos(b[2]), 0],
                        [a[0] * a[1] * np.cos(b[2]), a[1] ** 2, 0],
                        [0, 0, a[2] ** 2]])
        self.assertAlmostEqual((l.get_covariant_metric_tensor() - cmt).sum(), 0)
        cnt = np.linalg.inv(cmt)
        self.assertAlmostEqual((l.get_contravariant_metric_tensor() - cnt).sum(), 0)

    def test_d_equality(self):
        a = 1.0
        b = np.pi / 2
        l1 = br_mod.Lattice([a, a, a], [b, b, b])
        l2 = br_mod.Lattice([a, a, a], [b, b, b])
        self.assertEqual(l1, l2)

    def test_e_duals(self):
        d = br_mod.Lattice((1, 1, 1), np.array([1, 1, 1]) * np.pi / 2, real_space=True)
        r = br_mod.Lattice(np.array([1, 1, 1]) * np.pi * 2, np.array([1, 1, 1]) * np.pi / 2, real_space=False)
        self.assertEqual(d, r)

    def test_f_basis(self):
        basis = br_mod.Basis(self.cmo_positions, self.cmo_types)
        generators = br_mod.Symmetry(self.cmo_rotations, self.cmo_translations)
        symmetry = generators.generate()

        # Without specifying snap_to_symmetry the basis positions are as specified
        lat = br_mod.Lattice(self.cmo_vectors, symmetry, basis, snap_to_symmetry=False)
        self.assertTrue(np.allclose(lat.basis.positions, self.cmo_positions))

        # but specifying snap_to_symmetry=True forces the positions to take on their symmetry averages
        lat = br_mod.Lattice(self.cmo_vectors, symmetry, basis, snap_to_symmetry=True)
        self.assertTrue(np.allclose(lat.basis.positions, self.cmo_avg_positions))

    def test_g_wrapped_constructor(self):
        # build tuples here just to avoid typing the same thing multiple times
        bt = self.cmo_positions, self.cmo_types
        st = self.cmo_rotations, self.cmo_translations

        # Providing a Basis with no symmetry information should raise a ValueError
        self.assertRaises(ValueError, br_py.Lattice, self.cmo_vectors, basis=bt)

        # Providing both a spacegroup string (or strings) and Symmetry information should raise an error
        self.assertRaises(ValueError, br_py.Lattice, self.cmo_vectors, spacegroup='ITName', symmetry=st)
        self.assertRaises(ValueError, br_py.Lattice, self.cmo_vectors, spacegroup=('HMSymbol', 'HMChoice'), symmetry=st)

        # Providing Symmetry and Basis as positional arguments raises an error
        self.assertRaises(TypeError, br_py.Lattice, self.cmo_vectors, br_mod.Symmetry(*st), br_mod.Basis(*bt))

        # Until we can stop using Python 3.6 and 3.7 the first argument *can not* be position-only
        # # The single required argument can not be provided as a keyword
        # self.assertRaises(TypeError, br_py.Lattice, values=self.cmo_vectors)

        # And none of kwargs can have the same name 'values'
        self.assertRaises(TypeError, br_py.Lattice, self.cmo_vectors, values=True)

        # Providing Symmetry and Basis as keyword arguments works
        lat = br_py.Lattice(self.cmo_vectors, symmetry=br_mod.Symmetry(*st), basis=br_mod.Basis(*bt))
        self.assertTrue(np.allclose(lat.basis.positions, self.cmo_positions))

        # The objects can be created for us from tuple arguments if we wish:
        lat = br_py.Lattice(self.cmo_vectors, symmetry=st, basis=bt)
        self.assertTrue(np.allclose(lat.basis.positions, self.cmo_positions))

        # and of course, we can pass extra keywords (and their order doesn't matter)
        lat = br_py.Lattice(self.cmo_vectors, basis=bt, snap_to_symmetry=False, symmetry=st)
        self.assertTrue(np.allclose(lat.basis.positions, self.cmo_avg_positions))


if __name__ == '__main__':
    unittest.main()



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
import shlex
import subprocess
sys.path.insert(0, os.path.abspath('.'))

# -- Project information -----------------------------------------------------

project = 'brille'
copyright = '2020, Gregory Tucker'
author = 'Gregory Tucker'
import brille._brille as brille_module
version = brille_module.__version__ # just the 'short' version
release = brille_module.version # the 'full' version information


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.intersphinx',
    'sphinx.ext.autosummary',
    'sphinx.ext.napoleon',
    'breathe',
    'sphinxcontrib.katex',
    'sphinxcontrib.tikz',
    #'exhale',
]

# Useful mappings: https://gist.github.com/bskinn/0e164963428d4b51017cebdb6cda5209
intersphinx_mapping = {
  'euphonic': ('https://euphonic.readthedocs.io/en/stable/', None),
  'brilleu': ('https://brille.github.io/brilleu/latest/', None),
  'numpy': ('https://numpy.org/doc/stable/', None),
}

# Some :math:`[LaTeX]` directives insert '\r' into the string passed to katex?
# This raises an error with the version specified in the Docker image's
# sphinxcontrib.katex, 0.11.1, switching to v0.15.0 turns the error into
# a warning about 'LaTeX-incompatible input' 
# Updating the Docker image might be possible, but the exhale/breathe/sphinx
# problem encountered outside of the Docker build would need to be avoided.
katex_css_path = 'https://cdn.jsdelivr.net/npm/katex@0.15.0/dist/katex.min.css'
katex_js_path = 'https://cdn.jsdelivr.net/npm/katex@0.15.0/dist/katex.min.js'
katex_autorender_path = 'https://cdn.jsdelivr.net/npm/katex@0.15.0/dist/contrib/auto-render.min.js'

napoleon_use_ivar = True
napoleon_use_param = False
napoleon_use_admonition_for_notes = True

tikz_proc_suite = 'pdf2svg'  # We are building exclusively in our own container, no ReadTheDocs ghostscript restriction

breathe_projects = {'brille' : '_build/doxygenxml/'}
breathe_default_project = 'brille'
breathe_domain_by_extension = {'h': 'cpp', 'h': 'hpp', 'h': 'tpp'}

# # setup the exhale extension
# exhale_args = {
#     # required arguments first:
#     "containmentFolder": "./api",
#     "rootFileName": "library_root.rst",
#     "rootFileTitle": "C++ Library API",
#     "doxygenStripFromPath": "..",
#     # Suggested optional arguments
#     "createTreeView": True,
#     # TIP: if using the sphinx-bootstrap-theme, you also need
#     # "treeViewIsBootstrap": True,
#     "exhaleExecutesDoxygen": True,
#     #"exhaleSilentDoxygen": True,
#     "exhaleUseDoxyfile": True,
#     # "exhaleDoxygenStdin": "INPUT = ../wrap/", # "INPUT = ../src/",
# }

autosummary_generate = True

autoclass_content = 'both'

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

source_suffice = '.rst'
master_doc = 'index'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
#exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']
exclude_patterns = ['Thumbs.db', '.DS_Store']


# -- Options for HTML output -------------------------------------------------
# # following github.com/pybind/pybind11/blob/stable/docs/conf.py:
# this_is_rtd = os.environ.get('READTHEDOCS',None) == 'True'
# if this_is_rtd:
#     html_context = {'css_files' :[
#         '//media.readthedocs.org/css/sphinx_rtd_theme.css',
#         '//media.readthedocs.org/css/readthedocs-doc-embed.css',
#         '_static/theme_overrides.css'
#     ]}
# else:
#     import sphinx_rtd_theme
#     html_theme = 'sphinx_rtd_theme'
#     html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]
#     html_context = {'css_files': ['_static/theme_overrides.css']}

# Alternatively, just specify a theme:
html_theme = 'sphinx_rtd_theme'
html_logo = '../brille.svg'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

def call_and_check(command, **kwds):
    from subprocess import call, CalledProcessError
    try:
        retcode = call(command, **kwds)
        if retcode < 0:
            sys.stderr.write('{} error code: {}\n'.format(call, -retcode))
    except (CalledProcessError, OSError) as e:
        sys.stderr.write('{} execution failed: {}\n'.format(call, e))

# again, following from github.com/pybind/pybind11/blob/stable/docs/conf.py:
def generate_doxygen_xml(app):
    build_dir = os.path.join(app.confdir, '_build')
    if not os.path.exists(build_dir):
        os.mkdir(build_dir)

    call_and_check('doxygen', cwd=app.confdir)
    call_and_check(['breathe-apidoc','--output-dir=_build/breathe','-f',
                    '-g','class,namespace','_build/doxygenxml/'], cwd=app.confdir)

def setup(app):
    """Add hook for building Doxygen xml when needed"""
    app.connect("builder-inited", generate_doxygen_xml)



# Copyright © 2020 Greg Tucker <greg.tucker@stfc.ac.uk>
#
# This file is part of brille.
#
# brille is free software: you can redistribute it and/or modify it under the
# terms of the GNU Affero General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# brille is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#
# See the GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with brille. If not, see <https://www.gnu.org/licenses/>.

"""Python module :py:mod:`brille`
=================================

This module provides access to the C++ brille library which can be used to
interact with spacegroup and pointgroup symmetry operations, to determine the
first Brillouin zone for a given real space crystallographic lattice, to
find *an* irreducible polyhedron from the first Brillouin zone and the
pointgroup operations of a spacegroup, to construct polyhedron-filling connected
point networks, and to perform linear-interpolation of user-provided data for
any reciprocal space point at a symmetry-equivalent position within the
connected point network.

.. currentmodule:: brille

.. autosummary::
    :toctree: _generate
"""
from .bound import *
from .lattice import Lattice
from . import utils

try:
    from . import plotting
except ModuleNotFoundError:
    # Build servers don't have Matplotlib installed; plotting not tested
    pass



import os
import sys
try:
    from importlib.util import find_spec
except ImportError:
    raise Exception("This test requires importlib.util (and Python3)")
from timeit import default_timer
import numpy as np, matplotlib.pyplot as pp

for module in [('euphonic', 'data', 'interpolation'), ('brille', 'euphonic')]:
    for i in range(len(module)):
        check = '.'.join(np.array(module)[0:i+1])
        if find_spec(check) is None:
            raise Exception('Required module {} not found'.format(check))
from euphonic.data.interpolation import InterpolationData
from brille.euphonic import BrEu
import brille

def load_interpolation_data(named):
    """Load a data file from the repository tests folder

    Note that this only works if the package is installed via
        python setup.py develop [--user]
    which builds the C++ library and installs it inside of the repository
    structure with simlinks in the usual package location(s).
    """
    test_spec = find_spec('brille')
    brilleroot = test_spec.submodule_search_locations[0]
    seed = os.path.join(brilleroot,'..','tests',named)
    return InterpolationData(seed)

class timer:
    def __init__(self):
        self.tic()
    def tic(self):
        self.times_called = 0
        self.start_time = default_timer()
        self.elapsed_time = 0
    def toc(self):
        self.times_called += 1
        self.elapsed_time =  default_timer()-self.start_time
        return self.elapsed_time
    def elapsed(self):
        return self.elapsed_time
    def average(self):
        return self.elapsed_time/self.times_called
    def uncertainty(self):
        return np.sqrt(self.average())/self.times_called
    def relative_uncertainty(self):
        if self.times_called < 1:
            return np.inf
        return 1/np.sqrt(self.times_called*self.elapsed_time)
tictoc = timer()

def cen2corner(Z):
    dX = np.diff(Z, axis=0)/2
    Z = np.concatenate((
        Z[:-1]-dX,
        Z[np.newaxis, -1]-dX[np.newaxis, -1],
        Z[np.newaxis, -1]+dX[np.newaxis, -1]), axis=0)
    dY = np.diff(Z, axis=1)/2
    Z = np.concatenate((
        Z[:, :-1]-dY,
        Z[:, np.newaxis, -1]-dY[:, np.newaxis, -1],
        Z[:, np.newaxis, -1]+dY[:, np.newaxis, -1]), axis=1)
    return Z

def pcolormesh(X,Y,Z, *args,**kwargs):
    return pp.pcolormesh(cen2corner(X), cen2corner(Y), Z, *args, **kwargs)



dlat = brille.Direct((5.,5.,5.), (90.,90.,90.), 525)
bz = brille.BrillouinZone(dlat.star)

# max_sizes = 1/np.arange(100,5000,500)
max_sizes = 2**np.arange(-12,-2.01,0.5)
max_sizes = 1/(10**np.linspace(6,12,5))**(1/3)

print('pybind object creation times:')
it = np.nditer([max_sizes, None, None, None])
for s, c, e, n in it:
    tictoc.tic()
    while not (tictoc.elapsed() > 10 or tictoc.relative_uncertainty() < 0.01):
        mesh = brille.BZMeshQ(bz, max_size=s, lattice_ratio=2.)
        tictoc.toc()
    c[...] = tictoc.average()
    e[...] = tictoc.uncertainty()
    n[...] = mesh.rlu.shape[0]
    print('size {:6.4f}: <time>={:5.3f}({:5.3f}) sum(time)={:5.2f} sec'.format(s,c,e,tictoc.elapsed()))
creation = it.operands[1]
creation_uncertainty = it.operands[2]
n_vertices = it.operands[3]

pp.figure()
pp.errorbar(n_vertices**2, 1000*creation, yerr=1000*creation_uncertainty, marker='o')
pp.xlabel(r'$N_v^2$')
pp.ylabel(r'BZMeshQ object creation time / ms')

pp.figure()
pp.errorbar(n_vertices, 1000*creation, yerr=1000*creation_uncertainty, marker='o')
pp.xlabel(r'$N_v$')
pp.ylabel(r'BZMeshQ object creation time / ms')

pp.show()



import os
import sys
try:
    from importlib.util import find_spec
except ImportError:
    raise Exception("This test requires importlib.util (and Python3)")
from timeit import default_timer
import numpy as np, matplotlib.pyplot as pp

for module in [('euphonic', 'data', 'interpolation'), ('brille', 'euphonic')]:
    for i in range(len(module)):
        check = '.'.join(np.array(module)[0:i+1])
        if find_spec(check) is None:
            raise Exception('Required module {} not found'.format(check))
from euphonic.data.interpolation import InterpolationData
from brille.euphonic import BrEu
import brille

def load_interpolation_data(named):
    """Load a data file from the repository tests folder

    Note that this only works if the package is installed via
        python setup.py develop [--user]
    which builds the C++ library and installs it inside of the repository
    structure with simlinks in the usual package location(s).
    """
    test_spec = find_spec('brille')
    brilleroot = test_spec.submodule_search_locations[0]
    seed = os.path.join(brilleroot,'..','tests',named)
    return InterpolationData(seed)

class timer:
    def __init__(self):
        self.tic()
    def tic(self):
        self.times_called = 0
        self.start_time = default_timer()
        self.elapsed_time = 0
    def toc(self):
        self.times_called += 1
        self.elapsed_time =  default_timer()-self.start_time
        return self.elapsed_time
    def elapsed(self):
        return self.elapsed_time
    def average(self):
        return self.elapsed_time/self.times_called
    def uncertainty(self):
        return np.sqrt(self.average())/self.times_called
    def relative_uncertainty(self):
        if self.times_called < 1:
            return np.inf
        return 1/np.sqrt(self.times_called*self.elapsed_time)
tictoc = timer()

def cen2corner(Z):
    dX = np.diff(Z, axis=0)/2
    Z = np.concatenate((
        Z[:-1]-dX,
        Z[np.newaxis, -1]-dX[np.newaxis, -1],
        Z[np.newaxis, -1]+dX[np.newaxis, -1]), axis=0)
    dY = np.diff(Z, axis=1)/2
    Z = np.concatenate((
        Z[:, :-1]-dY,
        Z[:, np.newaxis, -1]-dY[:, np.newaxis, -1],
        Z[:, np.newaxis, -1]+dY[:, np.newaxis, -1]), axis=1)
    return Z

def pcolormesh(X,Y,Z, *args,**kwargs):
    return pp.pcolormesh(cen2corner(X), cen2corner(Y), Z, *args, **kwargs)



dlat = brille.Direct((5.,5.,5.), (90.,90.,90.), 525)
bz = brille.BrillouinZone(dlat.star)

# max_sizes = 1/np.arange(100,5000,500)
# trellis_r = 1/(np.arange(1,30002,5000)**(1/3))
max_sizes = 1/np.arange(100, 4000, 500)
# trellis_r = 1/(np.arange(1, 15002, 5000)**(1/3))
trellis_r = 2**np.arange(-1, 2.1, 0.5)

smat, tmat = np.meshgrid(max_sizes, trellis_r)

# print('pybind object creation times:')
# it = np.nditer([smat, tmat, None, None])
# for s, t, c, e in it:
#     tictoc.tic()
#     # while not (tictoc.elapsed() > 10 or tictoc.relative_uncertainty() < 0.1):
#     while not (tictoc.elapsed() > 1 or tictoc.relative_uncertainty() < 0.5):
#         brille.BZMeshQ(bz, max_size=s, lattice_ratio=t)
#         tictoc.toc()
#     c[...] = tictoc.average()
#     e[...] = tictoc.uncertainty()
#     print('size {:6.4f} ratio {:6.4f}: <time>={:5.3f}({:5.3f}) sum(time)={:5.2f} sec'.format(s,t,c,e,tictoc.elapsed()))
# creation = it.operands[2]
# creation_uncertainty = it.operands[3]
#
# pp.figure()
# pcolormesh(smat, tmat, np.log10(creation))
# pp.xlabel(r'maximum tetrahedral volume / $\AA^{-3}$')
# pp.ylabel(r'location trellis spacing / maximum tetrahedral circumsphere radius')
#
# pp.figure()
# for i, value in enumerate(trellis_r):
#     pp.errorbar(1/max_sizes, creation[i], yerr=creation_uncertainty[i], marker='o', linestyle='-', label='{:8.6f}'.format(value))
# pp.xlabel(r'inverse maximum tetrahedral volume / $\AA^3$')
# pp.ylabel(r'BZMeshQ object creation time / s')
# pp.legend(title='[trellis spacing]/[maximum circumsphere radius]')
#
# pp.figure()
# for i, value in enumerate(max_sizes):
#     pp.errorbar(1/trellis_r**3, creation[:,i], yerr=creation_uncertainty[:,i], marker='o', linestyle='-', label='{:8.6f}'.format(value))
# pp.xlabel(r'[maximum tetrahedral circumsphere radius]/[trellis spacing]')
# pp.ylabel(r'BZMeshQ object creation time / s')
# pp.legend(title=r'maximum tetrahedral volume / $\AA^{-1}$')
#
# pp.show()

# Next investigate now choice of maximum tetrahedra size and trellis size
# effect point-interpolation time
qx, qy = np.mgrid[1-0.3:1+0.3:0.005, 1-0.3:1+0.3:0.005]

# inshape = qx.shape
# px = cen2corner(qx)
# py = cen2corner(qy)

qx = qx.reshape(qx.size, 1)
qy = qy.reshape(qy.size, 1)
qxyz = np.concatenate((qx, qy, 0*qx), axis=1)

nb = load_interpolation_data('nb')

interpolation_trellis_ratios = 2**np.arange(-3,1.51,0.1)

print('Interpolation of {} points:'.format(qx.size))
it = np.nditer([interpolation_trellis_ratios, None, None])
for tr, i, ie in it:
    se = BrEu(nb, mesh=True, max_size=0.01, lattice_ratio=tr)
    tictoc.tic()
    while not (tictoc.elapsed() > 5 or tictoc.relative_uncertainty() < 0.01):
    # while not (tictoc.elapsed() > 1 or tictoc.relative_uncertainty() < 0.5):
        se(qxyz)
        tictoc.toc()
    i[...] = tictoc.average()
    ie[...] = tictoc.uncertainty()
    print('ratio {:6.4f}: <time>={:5.3f}({:5.3f}) sum(time)={:5.2f} sec'.format(tr,i,ie,tictoc.elapsed()))
interpolation_time = it.operands[1]
interpolation_uncertainty = it.operands[2]

pp.figure()
pp.errorbar(interpolation_trellis_ratios, interpolation_time, yerr=interpolation_uncertainty, marker='o', linestyle='-')
pp.xlabel(r'trellis spacing / maximum-circumsphere radius')
pp.ylabel('interpolation of {} points / s'.format(qx.size))

pp.figure()
pp.errorbar((4/3*np.pi)/interpolation_trellis_ratios**3, interpolation_time, yerr=interpolation_uncertainty, marker='o', linestyle='-')
pp.xlabel(r'$\rho_t$ / maximum-circumsphere')
pp.ylabel('interpolation of {} points / s'.format(qx.size))

pp.show()





def get_client_name():
    from shutil import which
    # python_on_whales supported clients, sorted by our preference
    commands = ('podman', 'docker', 'nerdctl')
    for command in commands:
        if which(command) is not None:
            return command
    raise RuntimeError("One of [docker, podman, nerdctl] required to use virtual-image based builder")


def get_client():
    from python_on_whales import DockerClient
    return DockerClient(client_call=[get_client_name()])


def get_image(client, image=None):
    if image is None:
        image = 'quay.io/pypa/manylinux2014_x86_64'
    quiet = client.image.exists(image)
    return client.image.pull(image, quiet=quiet)


def get_folder(name):
    from pathlib import Path
    folder = Path(__file__).parent.joinpath(name)
    if not folder.exists():
        folder.mkdir()
    return folder


def get_volumes(client, folders: dict):
    from pathlib import Path
    # ensure the cmake directory exists, relative to this file path
    folders['source'] = Path(__file__).parent.parent
    for key in [k for k, v in folders.items() if v is None]:
        folders[key] = get_folder(key)
    volumes = [[v, f"/{k}"] for k, v in folders.items()]
    name = client.client_config.client_call[0]
    if 'podman' in name:
        for volume in volumes:
            volume.append('Z')
    # the python_on_whales volumes keyword requires list(tuple(...)):    
    return [tuple(volume) for volume in volumes]


def write_entrypoint(entrypoint_string, folder, variant=None, sub=None):
    from os import cpu_count
    if variant is None:
        variant = 'cp310-cp310'
    if sub is None:
        sub = {}
    path = f"/opt/python/{variant}/bin"
    python = f"{path}/python" 
    count = cpu_count()
    filepath = folder.joinpath('entrypoint.sh')
    with open(filepath, 'w') as file:
        file.writelines(entrypoint_string.format(path=path, python=python, count=1 if count is None else count >> 1, **sub))





#!/usr/bin/env python3
"""Run tests of Brillouin zone creation."""
import unittest
import numpy as np
from importlib.util import find_spec
import brille as br_py


HASMPL = find_spec('matplotlib') is not None
HASMPL &= find_spec('mpl_toolkits') is not None
# protect against trying to load a submodule of a non-existant module
if HASMPL:
    HASMPL &= find_spec('mpl_toolkits.mplot3d') is not None
if HASMPL:
    import matplotlib.pyplot as pp
    from mpl_toolkits.mplot3d.axes3d import Axes3D


# pylint: disable=c0103
def plot_points(x, title=''):
    """Plot the 3D points with a given title."""
    if HASMPL:
        fig = pp.figure()
        ax = Axes3D(fig)
        ax.scatter(x[:, 0], x[:, 1], x[:, 2], s=10)
        ax.set_title(title)
        pp.show()


def plot_points_with_lines(x, y, title=''):
    """Plot 3D points and lines with a given title."""
    if HASMPL:
        fig = pp.figure()
        ax = Axes3D(fig)
        ax.plot(y[:, 0], y[:, 1], y[:, 2])
        ax.scatter(x[:, 0], x[:, 1], x[:, 2], s=10)
        ax.set_title(title)
        pp.show()


def plot_2d_points(x, title=''):
    """Plot 2D points and lines with a given title."""
    if HASMPL:
        ax = pp.axes()
        ax.scatter(x[:, 0], x[:, 1], s=10)
        ax.set_title(title)
        pp.show()


def make_dr(a, b, c, al=np.pi / 2, be=np.pi / 2, ga=np.pi / 2):
    """Make a Direct and Reciprocal lattice from Direct lattice parameters."""
    return br_py.Lattice(([a, b, c], [al, be, ga]))


def norm(x):
    return np.sqrt(np.dot(x, x))


def vector_lists_match(A, B):
    pA, pB = np.where(np.isclose(np.abs(A[:, np.newaxis, :] - B).sum(axis=2), 0.))
    if pA.size != A.shape[0] or pB.size != B.shape[0]:
        return False
    return np.allclose(A[pA], B[pB])


class BrillouinZone(unittest.TestCase):
    def test_a_init_unit_cube(self):
        # instantiate a cubic lattice with unit-length vectors
        # and its reciprocal lattice, still cubic with 2π-length vectors
        lat = make_dr(1, 1, 1)
        # creating a BrillouinZone objects requires that we pass the lattice
        # its first Brillouin zone is a cube in reciprocal space
        bz = br_py.BrillouinZone(lat)
        # with six faces, defined by: (00̄1),(0̄10),(̄100),(100),(010),(001)
        p = bz.points
        self.assertEqual(p.ndim, 2)
        self.assertEqual(p.shape[0], 6)  # and there is one for each of the six faces
        self.assertEqual(p.shape[1], 3)  # the face vectors are 3-vectors
        n = bz.normals
        self.assertEqual(n.ndim, 2)
        self.assertEqual(n.shape[0], 6)  # and there is one for each of the six faces
        self.assertEqual(n.shape[1], 3)  # the face normals are 3-vectors
        n_dot_p = np.array([np.dot(x / norm(x), y / norm(y)) for x, y in zip(n, p)])
        self.assertTrue((n_dot_p == 1.0).all())

        expected = np.array([[-1, 0, 0], [0, -1, 0], [0, 0, -1], [0, 0, 1], [0, 1, 0], [1, 0, 0]])
        self.assertTrue(vector_lists_match(p, expected / 2))
        self.assertTrue(vector_lists_match(np.array([x / norm(x) for x in n]), expected))
        #
        # the vertices of the first Brillouin zone are the 8 corners of the 1/2-unit cube:
        expected = np.array([[-1, -1, -1], [-1, -1, 1], [-1, 1, -1], [-1, 1, 1],
                             [1, -1, -1], [1, -1, 1], [1, 1, -1], [1, 1, 1]]) / 2
        verts = bz.vertices
        self.assertEqual(verts.ndim, 2)
        self.assertEqual(verts.shape[0], 8)
        self.assertEqual(verts.shape[1], 3)
        self.assertTrue(vector_lists_match(verts, expected))

    def test_b_isinside_unit_cube(self):
        lat = make_dr(1, 1, 1)
        bz = br_py.BrillouinZone(lat)
        # the first Brillouin zone of this unit-cube is bounded by
        # {(00̄1),(0̄10),(̄100),(100),(010),(001)}/2
        face_centres = np.array([[-1, 0, 0], [0, -1, 0], [0, 0, -1], [0, 0, 1], [0, 1, 0], [1, 0, 0]],
                                dtype='double') / 2
        self.assertTrue(np.all(bz.isinside(face_centres)))
        # including the vertices (since they are the corners of the zone)
        corners = np.array([[-1, -1, -1], [-1, -1, 1], [-1, 1, -1], [-1, 1, 1],
                            [1, -1, -1], [1, -1, 1], [1, 1, -1], [1, 1, 1]], dtype='double') / 2
        self.assertTrue(np.all(bz.isinside(corners)))
        # so all points with h, k, and l in the range [-0.5, 0.5] are in the zone
        Q = np.random.rand(100, 3) - 0.5  # this is a uniform distribution over [-0.5, 0.5) -- close enough
        self.assertTrue(np.all(bz.isinside(Q)))
        self.assertFalse(np.all(bz.isinside(Q + 5)))

    def test_c_moveinto_unit_cube(self):
        lat = make_dr(1, 1, 1)
        bz = br_py.BrillouinZone(lat)
        Q = (np.random.rand(100, 3) - 0.5) * 10  # this is a uniform distribution over [-5, 5)
        if np.all(bz.isinside(Q)):  # this is vanishingly-unlikely
            Q += 100.0
        self.assertFalse(np.all(bz.isinside(Q)))
        (q, tau) = bz.moveinto(Q)
        self.assertTrue(np.all(bz.isinside(q)))
        self.assertAlmostEqual(np.abs(Q - q - tau).sum(), 0)

    #
    #
    #
    def test_a_init_hexagonal(self):
        # instantiate a hexagonal lattice and its reciprocal lattice, still hexagonal
        lat = make_dr(3, 3, 9, np.pi / 2, np.pi / 2, np.pi * 2 / 3)
        # creating a BrillouinZone objects requires that we pass the lattice object
        bz = br_py.BrillouinZone(lat)
        # with eight faces, defined by: (̄100),(̄110),(0̄10),(001),(001),(010),(1̄10),(100)
        p = bz.points
        self.assertEqual(p.ndim, 2)
        self.assertEqual(p.shape[0], 8)  # and there is one for each of the eight faces
        self.assertEqual(p.shape[1], 3)  # the face vectors are 3-vectors
        n = bz.normals
        self.assertEqual(n.ndim, 2)
        self.assertEqual(n.shape[0], 8)  # and there is one for each of the six faces
        self.assertEqual(n.shape[1], 3)  # the face normals are 3-vectors
        n_dot_p = np.array([np.dot(x / norm(x), y / norm(y)) for x, y in zip(n, p)])
        self.assertTrue(np.allclose(n_dot_p, 1.))

        expected = np.array(
            [[-1, 0, 0], [-1, 1, 0], [0, -1, 0], [0, 0, -1], [0, 0, 1], [0, 1, 0], [1, -1, 0], [1, 0, 0]])
        self.assertTrue(vector_lists_match(p, expected / 2))
        n_compare = np.array([x / np.max(np.abs(x)) for x in n])
        self.assertTrue(vector_lists_match(n_compare, expected))
        self.assertTrue(np.all(bz.isinside(expected / 2)))
        #
        # the vertices of the first Brillouin zone are the 12 corners of the hexagonal-prism:
        expected = np.array([[-4, 2, -3], [-2, -2, -3], [-4, 2, 3], [-2, -2, 3], [-2, 4, -3], [-2, 4, 3],
                             [2, -4, -3], [2, -4, 3], [2, 2, -3], [4, -2, -3], [2, 2, 3], [4, -2, 3]]) / 6
        verts = bz.vertices
        self.assertEqual(verts.ndim, 2)
        self.assertEqual(verts.shape[0], 12)
        self.assertEqual(verts.shape[1], 3)
        self.assertTrue(vector_lists_match(verts, expected))
        self.assertTrue(np.all(bz.isinside(expected)))

    def test_b_isinside_hexagonal(self):
        lat = make_dr(3, 3, 9, np.pi / 2, np.pi / 2, np.pi * 2 / 3)
        bz = br_py.BrillouinZone(lat)
        # Q = (np.random.rand(1000, 3)-0.5) * 2 # this is a uniform distribution over [-1, 1)
        x = np.linspace(-1, 1, 100)
        X, Y, Z = np.meshgrid(x, x, 0)
        Q = np.stack((X.flatten(), Y.flatten(), Z.flatten()), axis=-1)
        Qin = bz.isinside(Q)
        B = lat.reciprocal_vectors
        X = np.stack([np.matmul(B, v) for v in Q[Qin, :]])
        # plot_2d_points(X)


    def test_c_moveinto_hexagonal(self):
        lat = make_dr(3, 3, 9, np.pi / 2, np.pi / 2, np.pi * 2 / 3)
        bz = br_py.BrillouinZone(lat)
        Q = (np.random.rand(100, 3) - 0.5) * 10  # this is a uniform distribution over [-5, 5)
        if np.all(bz.isinside(Q)):  # this is vanishingly-unlikely
            Q += 100.0
        self.assertFalse(np.all(bz.isinside(Q)))
        (q, tau) = bz.moveinto(Q)
        self.assertTrue(np.all(bz.isinside(q)))
        self.assertAlmostEqual(np.abs(Q - q - tau).sum(), 0)


    # This test can not work since I've removed the Hall-number symmetry specification constructor from Lattice
    # construction. ... but not from Spacegroup!
    def test_d_all_hallgroups(self):
        tested = 0
        failed = 0
        errored = 0
        failed_spg = []
        failed_ptg = []
        failed_lat = []
        failed_ratio = []
        errored_spg = []
        errored_ptg = []
        errored_lat = []
        errored_arg = []
        print()
        for i in range(1, 531):
            spacegroup = br_py.Spacegroup(i)
            pointgroup = br_py.Pointgroup(spacegroup.pointgroup_number)
            a, b, c, al, be, ga = 5, 5, 5, np.pi/2, np.pi/2, np.pi/2
            # nothing to do for cubic spacegroups
            if 'hexa' in pointgroup.holohedry:
                ga = 2 * np.pi / 3
            elif 'trig' in pointgroup.holohedry:
                if 'R' in spacegroup.choice:
                    al = be = ga = np.pi / 3
                else:  # 'H' setting or normally hexagonal
                    c, ga = 10, 2 * np.pi / 3
            elif 'tetr' in pointgroup.holohedry:
                c = 10
            elif 'orth' in pointgroup.holohedry:
                axperm = spacegroup.choice.replace('-', '')
                if 'cab' in axperm:
                    c, a, b = 5, 10, 15
                elif 'cba' in axperm:
                    c, b, a = 5, 10, 15
                elif 'bca' in axperm:
                    b, c, a = 5, 10, 15
                elif 'bac' in axperm:
                    b, a, c = 5, 10, 15
                elif 'acb' in axperm:
                    a, c, b = 5, 10, 15
                else:
                    a, b, c = 5, 10, 15
            elif 'mono' in pointgroup.holohedry:
                # continue # skip all monoclinic pointgroups for now
                if 'a' in spacegroup.choice:
                    a, al = 10, np.pi / 180 * (91 + 19 * np.random.rand())
                elif 'b' in spacegroup.choice:
                    b, be = 10, np.pi / 180 * (91 + 19 * np.random.rand())
                elif 'c' in spacegroup.choice:
                    c, ga = 10, np.pi / 180 * (91 + 19 * np.random.rand())
                else:
                    print("Monoclinic without 'a', 'b', or 'c' choice?? ", spacegroup.choice)
                    continue
            elif 'tric' in pointgroup.holohedry:
                ang = lambda: np.pi / 3 * (1 + np.random.rand())
                a, b, c, al, be, ga = 5, 10, 15, ang(), ang(), ang()

            lat = br_py.Lattice(([a, b, c], [al, be, ga]), spacegroup=spacegroup.hall_symbol)

            # print("Hall ", i, " ", dlat)
            # print(spacegroup,pointgroup)
            try:
                bz = br_py.BrillouinZone(lat)
                vol_bz = bz.polyhedron.volume
                vol_ir = bz.ir_polyhedron.volume
                tested += 1
                if not np.isclose(vol_ir, vol_bz / br_py.PointSymmetry(i).size):
                    # print(dlat,": ",vol_ir," != ",vol_bz/br_py.PointSymmetry(i).size)
                    failed += 1
                    failed_spg.append(spacegroup)
                    failed_ptg.append(pointgroup)
                    failed_lat.append(lat)
                    failed_ratio.append(vol_ir / vol_bz * br_py.PointSymmetry(i).size)
            except Exception as err:
                errored += 1
                errored_spg.append(spacegroup)
                errored_ptg.append(pointgroup)
                errored_lat.append(lat)
                errored_arg.append(err.args[0])

        if failed > 0:
            print("\nFailed to find irreducible Brillouin zone for", failed, "out of", tested, "(max 530) Hall groups")
            for spg, ptg, lat, rat in zip(failed_spg, failed_ptg, failed_lat, failed_ratio):
                print(spg, ptg, lat, rat)
        if errored > 0:
            print("\nException raised for", errored, "out of", tested, "(max 530) Hall Groups")
            for spg, ptg, lat, arg in zip(errored_spg, errored_ptg, errored_lat, errored_arg):
                print(arg)
                print(spg, ptg, lat)
        self.assertTrue(errored == 0)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# This file should not be renamed 'test*.py' until module/submodule loading is
# figured out from within a build directory.
# At the moment, if you've built the _brille binary but not installed it for
# python to find on its path, this script puts the empty (__import__.py only)
# brille 'package' on the python search path in order to load the euphonic
# submodule. When the submodule is loaded the empty 'package' gets loaded as
# well, and then the whole thing fails to run.
"""Run tests of BrEu and Euphonic."""
import os
import sys
import unittest
from euphonic.data.interpolation import InterpolationData
try:
    from importlib.util import find_spec
except ImportError:
    raise Exception("This test requires importlib.util (and Python3)")
import numpy as np

# Try to find the brille module:
# It might be in the current directory, or a sub-directory called, e.g., Debug
ADDPATH = os.getcwd()
if os.path.exists('Debug'):
    ADDPATH = os.path.join(ADDPATH, 'Debug')
sys.path.append(ADDPATH)
# Make sure we can load the binary module before the euphonic submodule tries:
if find_spec('brille') is not None and find_spec('brille._brille') is not None:
    import brille
elif find_spec('_brille') is not None:
    # pylint: disable=e0401
    import _brille as brille
else:
    raise Exception("Required brille module not found!")
# We need to find the pure-python submodule brille.euphonic:
sys.path.append(os.path.split(os.getcwd())[0])
if find_spec('brille') is not None and find_spec('brille.euphonic') is not None:
    from brille.euphonic import BrEu
else:
    raise Exception("Required brille.euphonic module not found!")


def load_interpolation_data(named):
    """Load a data file from the tests folder."""
    b = find_spec('brille')
    if b is None:
        bdir = '.' # punt! maybe we're in a build directory?
    else:
        bdir = b.submodule_search_locations[0]
    testdir = os.path.join(bdir, '..', 'tests')
    if not os.path.exists(testdir):
        raise Exception('Could not locate the tests directory')
    return InterpolationData(seedname=named, path=testdir)


def hermitian_product(v_0, v_1, first=None, last=None):
    """Find the hermitian product between two sets of vectors."""
    return np.dot(np.conj(v_0[first:last].flatten()), v_1[first:last].flatten())


class TestEuphonic(unittest.TestCase):
    """A TestCase object class to run tests of the BrEu object."""

    def test_a(self):
        """Test whether SimPhony gives the same result *at* grid points.

        This test should always pass, since the BZGridQcomplex object is filled
        by calculating eigen-energies and eigen-vectors using SimPhony.
        Unfortunately, due to rounding errors, this test can fail to run if an
        out-of-bounds interpolation point is required by brille.

        The moveinto keyword is what makes it possible to verify the
        "interpolation" result against the SimPhony result. This same keyword
        also opens up the possibility of rounding-errors leading to a runtime
        error of brille.
        With the moveinto keyword omitted (or set to True), out-of-zone grid
        points *will* likely have different polarization vectors compared to
        the SimPhony result as one interpolates at q=Q-τ and the other
        calculates at Q. It's unclear if this difference is important for
        intensity calculations.
        """
        i_data = load_interpolation_data('nb')
        symsim = BrEu(i_data, halfN=(2, 2, 2))

        q_rlu = symsim.grid.rlu
        int_freq, int_vecs = symsim.frqs_vecs(q_rlu, interpolate=True,
                                              moveinto=False)
        sim_freq, sim_vecs = symsim.frqs_vecs(q_rlu, interpolate=False)

        ad_freq = np.abs(int_freq - sim_freq).magnitude
        as_freq = np.abs(int_freq + sim_freq).magnitude
        # Check if |interpolated-simulated|/|interpolated+simulated|>1e-14
        # AND if |interpolated-simulated|>1e-14
        unequal_freq = (ad_freq > 1e-14*as_freq) * (ad_freq > 1e-14)
        self.assertFalse(unequal_freq.any())

        # The vectors only need to be equal up to an arbitrary phase
        # which is equivalent to saying that the inner product between
        # equal ion eigenvectors for each branch should have ||ϵ⋅ϵ||²≡1
        n_pt, n_br, n_io, n_d = int_vecs.shape
        int_vecs = int_vecs.reshape(n_pt*n_br*n_io, n_d)
        sim_vecs = sim_vecs.reshape(n_pt*n_br*n_io, n_d)
        product = [hermitian_product(x, y) for x, y in zip(int_vecs, sim_vecs)]

        self.assertTrue(np.isclose(np.abs(product), 1).all())

    def test_b(self):
        """Do something."""
        i_data = load_interpolation_data('nb')
        symsim = BrEu(i_data, halfN=(10, 10, 10))
        print(symsim.grid.centre_sort_perm())


if __name__ == '__main__':
    unittest.main()



# The compiled module is installed at the base of the package directory structure such that its properties
# can be included here. If the module is loaded, however, without being installed first these import statements will
# raise a ModuleNotFoundError, which prevents automated pre-installation testing from taking place.
try:
    from ._brille import __version__, version, AngleUnit, LengthUnit, Bravais, RotatesLike, NodeType
    from ._brille import real_space_tolerance, reciprocal_space_tolerance, ApproxConfig
    from ._brille import Basis, HallSymbol, Spacegroup, Pointgroup, Symmetry, PointSymmetry, PrimitiveTransform
    from ._brille import BrillouinZone, Polyhedron, LPolyhedron, SortingStatus
    from ._brille import BZTrellisQcc, BZTrellisQdc, BZTrellisQdd
    from ._brille import BZMeshQcc, BZMeshQdc, BZMeshQdd
    from ._brille import BZNestQcc, BZNestQdc, BZNestQdd

    # Store the grid types for use in, e.g., the plotting routines
    __grid_types__ = (
        BZTrellisQcc, BZTrellisQdc, BZTrellisQdd,
        BZMeshQcc, BZMeshQdc, BZMeshQdd,
        BZNestQcc, BZNestQdc, BZNestQdd,
    )
except ModuleNotFoundError:
    pass



import subprocess
import sys
import platform
from datetime import datetime

def version_info():
    try:
        git_revision = subprocess.check_output(["git", "rev-parse", "HEAD"]).decode("utf-8") .split("\n")[0]
        git_branch = subprocess.check_output(["git", "rev-parse","--abbrev-ref", "HEAD"]).decode("utf-8").split("\n")[0]
    except (subprocess.CalledProcessError, OSError):
        git_revision = ""
        git_branch = "non-git"

    def read_version():
        with open("VERSION") as f:
            return f.readline().strip()

    build_datetime = datetime.now().isoformat(timespec='minutes')
    version_number = read_version()
    hostname = platform.node()
    return git_revision, git_branch, build_datetime, version_number, hostname

def version_number():
    sys.stdout.write(version_info()[3])

if __name__ =="__main__":

    output_file = sys.argv[1]
    with open(output_file, "w") as fout:
        fout.write("""#ifndef BRILLE_VERSION_HPP_
#define BRILLE_VERSION_HPP_
//! \\file
namespace brille::version{{
    //! `brille` git repository revision information at build time
    auto constexpr git_revision = u8"{0}";
    //! `brille` git repository branch at build time
    auto constexpr git_branch = u8"{1}";
    //! build date and time in YYYY-MM-DDThh:mm format
    auto constexpr build_datetime = u8"{2}";
    //! `brille` version
    auto constexpr version_number = u8"{3}";
    //! hostname of the build machine
    auto constexpr build_hostname = u8"{4}";
}}
#endif
""".format(*version_info()))



# Build and test the python module using CIBuildWheel

ENTRYPOINT="""#!/bin/bash
set -eu
# !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# ! This file is auto-generated and will be overwritten !
# !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

function echo_run {{
    echo "-------------------------------------------------"
	echo "$" "$@"
    echo "-------------------------------------------------"
	"$@"
}}

export PATH={path}:$PATH
export CONAN_USER_HOME=/conan

# Configure the project, specifying which python interpreter and library to use
echo_run {python} -m pip wheel -w /build --no-deps /source

echo_run find /build -type f -iname "*-linux*.whl" -exec sh -c "auditwheel repair '{{}}' -w \$(dirname '{{}}')" \;

"""


def main():
    from config import get_client, get_image, get_volumes, get_folder, write_entrypoint
    from python_on_whales import exceptions

    client = get_client()
    image = get_image(client)
    folder = get_folder('wheelhouse')
    volumes = get_volumes(client, {'build': folder, 'conan': None})
    
    write_entrypoint(ENTRYPOINT, folder)
    try:
        result = client.run(image, ['sh', '/build/entrypoint.sh'], volumes=volumes, tty=True)
        print(result)
    except exceptions.DockerException as ex:
        raise RuntimeError()


if __name__ == '__main__':
    main()




import os
import numpy as np, matplotlib as mpl, matplotlib.pyplot as pp
import brille as b, brille.plotting as bp
from brilleu.brilleu import getBrillEuObj

images_dir = os.path.join(os.path.dirname(__file__), 'images')
def dirsavefig(dirname, filename, transparent=True, **kwds):
    if not os.path.isdir(dirname):
        os.makedirs(dirname)
    pp.savefig(os.path.join(dirname, filename), transparent=transparent, **kwds)

def centre_to_corner(X):
    dX = np.diff(X, axis=0)/2
    X = np.concatenate((
        X[:-1]-dX,
        X[np.newaxis, -1]-dX[np.newaxis, -1],
        X[np.newaxis, -1]+dX[np.newaxis, -1]), axis=0)
    dY = np.diff(X, axis=1)/2
    X = np.concatenate((
        X[:, :-1]-dY,
        X[:, np.newaxis, -1]-dY[:, np.newaxis, -1],
        X[:, np.newaxis, -1]+dY[:, np.newaxis, -1]), axis=1)
    return X

# use_c=False ensures the Euphonic C module is *not* used.
# but *need* to use the brille C++ module, and parallel=True ensures we do so with OpenMP
scattering_lengths = {'Na': 3.63, 'Cl': 9.577} # in fm
nacl = getBrillEuObj('NaCl', max_volume=1e-5, parallel=True, sort=True, scattering_lengths=scattering_lengths)

# Define a path through reciprocal space from (000) to (123)
xi = np.linspace(0, 1, 100)
qhkl = np.vstack((xi, 2*xi, 3*xi)).T
# Interpolate the mode energies at qhkl
w_q = nacl.w_q(qhkl)

bz = nacl.grid.BrillouinZone
# plot the first Brillouin zones for (000), (111), (022), and (113) and path
fig = pp.figure(figsize=pp.figaspect(2))
ax = fig.add_axes([0,0,1,1], projection='3d')
ax.view_init(5,300)
for g in [(0,0,0), (1,1,1), (0,2,2), (1,1,3)]:
    bp.plot(bz, units='rlu', irreducible=False, origin=g, edgecolor='gray', alpha=0, axs=ax, show=mpl.is_interactive())
bp.plot(qhkl, axs=ax, show=mpl.is_interactive())
pp.setp(ax,'xlim',[-1,2],'ylim',[-1,2],'zlim',[-1,4])
pp.setp(ax,'xlabel','h','ylabel','k','zlabel','l')
bbox = fig.bbox_inches.from_bounds(0.2,0.8,4.4,5.7)
dirsavefig(images_dir, 'nacl_123_path.png', bbox_inches=bbox)

# plot the Brillouin zone, path, irreducible path, and dispersion
irqhkl, G, rot, invrot = nacl.grid.BrillouinZone.ir_moveinto(qhkl)
fig = pp.figure(figsize=pp.figaspect(0.5))
ax0 = fig.add_subplot(1, 2, 1, projection='3d')
bp.plot(bz, units='rlu', axs=ax0, Q=irqhkl, show=mpl.is_interactive())
pp.setp(ax0,'xlabel','h','ylabel','k','zlabel','l')

ax1 = fig.add_subplot(1, 2, 2)
for y in w_q.T:
	ax1.plot(xi, y)
ax1.set_xlabel(r'$\xi$ in $\mathbf{Q}=(\xi\,2\xi\,3\xi)$')
ax1.set_ylabel(r'$\omega_i(\mathbf{Q})$ / meV')

fig.subplots_adjust(left=0,right=0.99,bottom=0.1,top=1)
dirsavefig(images_dir, 'nacl_123_disp.png')

# determine and plot the inelastic neutron scattering intesity along the path
mu, nu = np.mgrid[0:1:50j, 0.5:30:60j]
ξ = mu.reshape(mu.size,1)
En = nu.reshape(nu.size,1)
Qhkl = np.concatenate((ξ, 2*ξ, 3*ξ), axis=1)

brSQE = nacl(Qhkl, En, resfun='sho', param=1., T=1.).reshape(mu.shape)
euSQE = nacl(Qhkl, En, resfun='sho', param=1., T=1., interpolate=False).reshape(mu.shape)

mu = centre_to_corner(mu)
nu = centre_to_corner(nu)
fig = pp.figure(figsize=pp.figaspect(0.5))
ax2 = fig.add_axes([0.07,0.1,0.4,0.8])
ax2.pcolormesh(mu, nu, brSQE, shading='flat')
ax2.set_xlabel(r'$\xi$ in $\mathbf{Q}=(\xi\,2\xi\,3\xi)$')
ax2.set_ylabel(r'$E$/meV')
ax2.set_title('brille')

ax3 = fig.add_axes([0.55,0.1,0.4,0.8])
ax3.pcolormesh(mu, nu, euSQE, shading='flat')
ax3.set_xlabel(r'$\xi$ in $\mathbf{Q}=(\xi\,2\xi\,3\xi)$')
ax3.set_ylabel(r'$E$/meV')
ax3.set_title('euphonic')

dirsavefig(images_dir, 'nacl_123_sqw.png')



class BrillePrinter:
    def __init__(self, val):
        self.val = val

    def method(self, named, *args):
        arg_string = ",".join(f"{arg}" for arg in args) if len(args) else ""
        eval_string = f"(*({self.val.type}*)({self.val.address})).{named}({arg_string})"
        return gdb.parse_and_eval(eval_string)

    def wrap_to_string(self, string):
        return str(string)

    def _to_string(self):
        return ""

    def to_string(self):
        return self.wrap_to_string(self._to_string())


class BrilleLatticePrinter(BrillePrinter):
    def _to_string(self):
        return str(self.method('string_repr'))


class BrilleArrayPrinter(BrillePrinter):
    def _to_string(self):
        string = str(self.method('to_string')).replace(r"\n", "\n")
        return f"\n{string}"


class BrilleLatVecPrinter(BrillePrinter):
    def _to_string(self):
        array = BrilleArrayPrinter(self.val)
        lattice = BrilleLatticePrinter(self.val['lattice'])
        return f"\n{lattice._to_string()}\n{array._to_string()}"


def array_printer(val):
    if "Array" in str(val.type): return BrilleArrayPrinter(val)

def lattice_printer(val):
    if "Lattice" in  str(val.type): return BrilleLatticePrinter(val)

def latvec_printer(val):
    if "LatVeC" in str(val.type): return BrilleLatVecPrinter(val)

# gdb.pretty_printers.append(lattice_printer)
# gdb.pretty_printers.append(array_printer)
# gdb.pretty_printers.append(latvec_printer)


import os
import re
import sys
import pkgutil
from sysconfig import get_platform
from subprocess import check_output, check_call
from pathlib import Path

from setuptools import setup, Extension, find_packages
from setuptools.command.build_ext import build_ext

if pkgutil.find_loader('packaging') is None:
    from distutils.version import LooseVersion as Version
else:
    from packaging.version import Version

# We can use cmake provided from pip which (normally) gets installed at /bin
# Except that in the manylinux builds it's placed at /opt/python/[version]/bin/
# (as a symlink at least) which is *not* on the path.
# If cmake is a known module, import it and use it to tell us its binary directory
if pkgutil.find_loader('cmake') is not None:
    import cmake

    CMAKE_BIN = cmake.CMAKE_BIN_DIR + os.path.sep + 'cmake'
else:
    CMAKE_BIN = 'cmake'


def get_cmake():
    return CMAKE_BIN


# We want users to be able to specify to *not* use HDF5 for object IO.
# Disable HDF5 IO by passing `--no-hdf` when calling python setup.py.
USE_HDF5 = True


def is_vsc():
    platform = get_platform()
    return platform.startswith("win")


def is_mingw():
    platform = get_platform()
    return platform.startswith("mingw")


class CMakeExtension(Extension):
    def __init__(self, name, sourcedir=''):
        Extension.__init__(self, name, sources=[])
        self.sourcedir = os.path.abspath(sourcedir)

PACKAGE_ROOT = Path(__file__).absolute().parent

if PACKAGE_ROOT != Path(os.getcwd()):
    raise RuntimeError(f"{PACKAGE_ROOT} != {os.getcwd()}")

class CMakeBuild(build_ext):
    def run(self):
        try:
            out = check_output([get_cmake(), '--version'])
        except OSError:
            raise RuntimeError("CMake must be installed to build" +
                               " the following extensions: " +
                               ", ".join(e.name for e in self.extensions))

        rex = r'version\s*([\d.]+)'
        cmake_version = Version(re.search(rex, out.decode()).group(1))
        if cmake_version < Version('3.18.2'):
            raise RuntimeError("CMake >= 3.18.2 is required")

        for ext in self.extensions:
            self.build_extension(ext)

    def build_extension(self, ext):
        extdir = os.path.dirname(self.get_ext_fullpath(ext.name))
        extdir = os.path.abspath(extdir)
        cmake_args = []
        if is_vsc():
            if sys.maxsize > 2 ** 32:
                cmake_args += ['-A', 'x64']
            else:
                cmake_args += ['-A', 'Win32']

        if is_mingw():
            cmake_args += ['-G', 'Unix Makefiles']  # Must be two entries to work

        cmake_args += ['-DCMAKE_LIBRARY_OUTPUT_DIRECTORY=' + extdir,
                       '-DPYTHON_EXECUTABLE=' + sys.executable]

        cfg = 'Debug' if self.debug else 'Release'
        # cfg = 'Debug' if self.debug else 'RelWithDebInfo'
        build_args = ['--config', cfg, '--target', '_brille']

        # make sure all library files end up in one place
        cmake_args += ["-DCMAKE_BUILD_WITH_INSTALL_RPATH=TRUE"]
        cmake_args += ["-DCMAKE_INSTALL_RPATH={}".format("$ORIGIN")]

        if not USE_HDF5:
            cmake_args += ["-DBRILLE_HDF5=FALSE"]

        if is_vsc():
            cmake_lib_out_dir = '-DCMAKE_LIBRARY_OUTPUT_DIRECTORY_{}={}'
            cmake_args += [cmake_lib_out_dir.format(cfg.upper(), extdir)]
            cmake_args += ['-DCMAKE_BUILD_TYPE=' + cfg]
            build_args += ['--', '/m:4']
        else:
            cmake_args += ['-DCMAKE_BUILD_TYPE=' + cfg]
            build_args += ['--', '-j']

        if not os.path.exists(self.build_temp):
            os.makedirs(self.build_temp)
        check_call([get_cmake(), str(PACKAGE_ROOT)] + cmake_args, cwd=self.build_temp)
        check_call([get_cmake(), '--build', '.', '--target', "_brille"] + build_args, cwd=self.build_temp)


with open(PACKAGE_ROOT.joinpath('README.md'), 'r') as fh:
    LONG_DESCRIPTION = fh.read()

if "--use-hdf5" in sys.argv:
    USE_HDF5 = True
    sys.argv.remove("--use-hdf5")
if "--no-hdf5" in sys.argv:
    USE_HDF5 = False
    sys.argv.remove("--no-hdf5")

setup(
    name='brille',
    author='Greg Tucker',
    author_email='gregory.tucker@ess.eu',
    description='Irreducible Brillouin zone symmetry and interpolation.',
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    ext_modules=[CMakeExtension('brille._brille')],
    packages=find_packages(str(PACKAGE_ROOT)),
    extras_require={'plotting': ['matplotlib>=2.2.0', ], 'vis': ['pyglet>=1.5.27', 'vispy>=0.12.1', ]},
    cmdclass=dict(build_ext=CMakeBuild),
    url="https://github.com/brille/brille",
    zip_safe=False,
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)",
        "Operating System :: Microsoft :: Windows :: Windows 10",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: C++",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering :: Physics",
    ]
)



#!/usr/bin/env python

# Obtained from https://github.com/mosra/corrade/tree/master/src/acme
# with minor modifications to suit brille
#       - UTF-8 encoded output files
#       - 0+ space characters before 'Copyright ©'
#       - search all include paths for 'local' header files
#       - use rstrip() to remove trailing carriage return character from regular
#         expression capture 'value' on Windows (which strips \n but not \r)
#
# Used by CMake make single header suitable for install from brille.h template.

#
#   This file is part of Corrade.
#
#   Copyright © 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015, 2016,
#               2017, 2018, 2019, 2020 Vladimír Vondruš <mosra@centrum.cz>
#
#   Permission is hereby granted, free of charge, to any person obtaining a
#   copy of this software and associated documentation files (the "Software"),
#   to deal in the Software without restriction, including without limitation
#   the rights to use, copy, modify, merge, publish, distribute, sublicense,
#   and/or sell copies of the Software, and to permit persons to whom the
#   Software is furnished to do so, subject to the following conditions:
#
#   The above copyright notice and this permission notice shall be included
#   in all copies or substantial portions of the Software.
#
#   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
#   THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#   FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#   DEALINGS IN THE SOFTWARE.
#

import argparse
import codecs
import re
import os
import logging
import subprocess

from typing import List, Tuple

alone_in_parentheses_rx = re.compile(r'\((?P<inside>0|1|!?defined\((?P<name>[^)]+)\))\)')
zero_and_something_rx = re.compile(r'0 && (0|1|!?defined\([^)]+\)|!?\()')
something_and_zero_rx = re.compile(r'(0|1|!?defined\([^)]+\)|\)) && 0')
one_or_something_rx = re.compile(r'1 \|\| (0|1|!?defined\([^)]+\)|!?\()')
something_or_one_rx = re.compile(r'(0|1|!?defined\([^)]+\)|\)) \|\| 1')
alone_defined_rx = re.compile(r'^(?P<not>!)?defined\((?P<name>[^)]+)\)$')

def normalize_expression(expression) -> Tuple[str, str]:
    match = alone_defined_rx.match(expression)
    if match:
        return 'ifndef' if match.group('not') else 'ifdef', match.group('name')
    return 'if', expression

def simplify_expression(what, expression, forced_defines = {}):
    assert what in ['if', 'ifdef', 'ifndef', 'elif']

    # Denormalize the shorthand expression
    if what == 'ifdef':
        what, expression = 'if', 'defined({})'.format(expression)
    if what == 'ifndef':
        what, expression = 'if', '!defined({})'.format(expression)

    # Go through all forced defines and replace them with their values
    for name, enabled in forced_defines.items():
        if name not in expression: continue

        expression = expression.replace('defined({})'.format(name), '1' if enabled else '0')

    # Now comes the ugly part, simplify as long as there's something
    modified = True
    while modified:
        modified = False # Nothing done in this round yet

        # Basic replacements
        for find, replace in [('!0', '1'),
                              ('!1', '0'),
                              # All these have to be after `!0` / `!1` so e.g.
                              # `!0 || b` doesn't get # simplified to `!b`
                              ('0 || ', ''),
                              (' || 0', ''),
                              ('1 && ', ''),
                              (' && 1', '')]:
            if expression.find(find) != -1:
                expression = expression.replace(find, replace)
                modified = True

        # defined() / 0 / 1 alone in parentheses, remove parentheses
        match = alone_in_parentheses_rx.search(expression)
        if match:
            expression = expression[:match.start()] + match.group('inside') + expression[match.end():]
            modified = True

        # If the above replacements modified something, restart to avoid e.g.
        # `!0 && (something)` getting reduced by the "zero and something" rule
        # below
        if modified: continue

        # 0 && something or something && 0, replace with just 0
        # 1 || something or something || 1, replace with just 1
        for rexp, repl in [(zero_and_something_rx, '0'), (something_and_zero_rx, '0'),
                           (one_or_something_rx, '1'), (something_or_one_rx, '1')]:
            match = rexp.search(expression)
            if match:
                # The "something" is a parenthesised expression after, go to
                # the right and eat everything in matching parentheses
                if match.group(1) in ['(', '!(']:
                    level = 1
                    for i in range(match.end(), len(expression)):
                        c = expression[i]
                        if c == '(': level = level + 1
                        elif c == ')':
                            level = level - 1
                            assert level >= 0
                            if level == 0:
                                expression = expression[:match.start()] + repl + expression[i + 1:]
                                break

                # The "something" is a parenthesised expression before, go to
                # the left and eat everything in matching parentheses
                elif match.group(1) == ')':
                    level = 1
                    for i in range(match.start(), 0, -1):
                        c = expression[i - 1]
                        if c == ')': level = level + 1
                        elif c == '(':
                            level = level - 1
                            assert level >= 0
                            if level == 0:
                                # Eat also the negation, if there
                                if i > 1 and expression[i - 2] == '!':
                                    i = i - 1
                                expression = expression[:i - 1] + repl + expression[match.end():]
                                break

                # Simple version
                else:
                    expression = expression[:match.start()] + repl + expression[match.end():]

                modified = True

    if expression == '0':
        result = False
    elif expression == '1':
        result = True
    else:
        result = None # keep it

    # Renormalize back to ifdef if the expression is simple enough
    if what == 'if': what, expression = normalize_expression(expression)

    return result, what, expression

def sort_includes(includes: List[str]) -> List[str]:
    system, local = [], []
    for i in sorted(includes):
        if i.startswith('#include <'): system += [i]
        elif i.startswith('#include "'): local += [i]
        else: assert False # pragma: no cover
    return (system + ["\n"] + local) if local and system else (system + local)

copyright_email_rx = re.compile(r'<[^@]+@[^>]+>')
copyright_year_rx = re.compile(r'\d\d\d\d')

def sort_copyrights(copyrights: List[str]) -> List[str]:
    email_dict = {}
    output = []

    # Go through all copyright, the oldest first
    for copyright in sorted(copyrights):
        match = copyright_email_rx.search(copyright)
        assert match
        email = match.group(0)

        # If the e-mail is not in the dict yet, add it there with all years it
        # covers
        years = set([int(i) for i in copyright_year_rx.findall(copyright)])
        if email not in email_dict:
            email_dict[email] = (years, len(output))
            output += [copyright]

        # If given mail is in the dict already, check that all years are there
        # as well
        else:
            # If all years are there, nothing left to do
            if years <= email_dict[email][0]: continue

            # If we have a superset of the years, use this copyright line
            # instead
            if years > email_dict[email][0]:
                output[email_dict[email][1]] = copyright
                email_dict[email] = (years, email_dict[email][1])
                continue

            # Otherwise complain
            raise ValueError("First copyright found for {} is missing years {}, supply an explicit combined copyright instead".format(email, repr(years - email_dict[email][0])))

    return output

include_rx = re.compile(r'^(?P<include>#include (?P<quote>["<])(?P<file>[^">]+)[">]).*?$')
preprocessor_rx = re.compile(r'^(?P<indent>\s*)#(?P<what>ifdef|ifndef|if|else|elif|endif)\s*(?P<value>[^\n]*?[^\s]?)(?P<comment>\s*/[/*].*)?$')
define_rx = re.compile(r'\s*#(?P<what>define|undef) (?P<name>[^\s]+)\s*$')
linecomment_rx = re.compile(r'^\s*(/\*.*\*/|//.*)?\s*$')
copyright_rc = re.compile(r'^\s*Copyright © \d{4}.+$')
blockcomment_start_rx = re.compile(r'^\s*/\*.*\s*$')
blockcomment_end_rx = re.compile(r'^\s*.*\*/\s*$')
acme_pragma_rx = re.compile(r'^#pragma\s+ACME\s+(?P<what>[^\s]+)\s*(?P<value>[^\s]?.*)\s*$')
tpp_filename = re.compile("""\.tpp$""")

def acme(toplevel_file, output) -> List[str]:
    base_directory = os.path.dirname(toplevel_file)

    write_comments = True
    paths = []
    local_include_prefixes = []
    local_includes_noexpand = set()
    all_includes = set()
    new_includes = []
    copyrights = set()
    parsed_files = set()
    forced_defines = {}
    revision_commands = {}
    stats_commands = {}
    def parse(file, level):
        nonlocal write_comments, paths, local_include_prefixes, all_includes, new_includes, copyrights, parsed_files, forced_defines, revision_commands, stats_commands

        logging.info("%sParsing file %s...", ' '*level, file)

        # Mark the file as parsed. Doing it at the beginning so accidental
        # circular includes are caught.
        parsed_files.add(file)

        # Output line buffer + line buffer for includes (which need to be
        # inserted before the actual output, but with their order preserved)
        includes_out = []
        out = []
        in_comment = False
        comment_buffer = []
        multiline_copyright: str = None
        # Preprocessor branch stack. First element of each item is True if
        # this is a real node, False if this is a dummy for an #elif. The leaf
        # node is never False, all False nodes get popped at an #endif. Second
        # element is True if contents of current branch should be printed,
        # False if not and None if the preprocessor branching should be kept
        # verbatim. There's always at least one element, using a two-element
        # list instead of a tuple so I can modify the entries
        branch_stack = [[True, True, 0]]

        bline: bytes
        with open(file, 'rb') as f:
            # We don't handle BOMs, but let the user know that clearly
            bom = f.read(len(codecs.BOM_UTF8))
            if bom == codecs.BOM_UTF8:
                raise ValueError("{} contains a BOM, this is ABSOLUTELY unacceptable".format(file))
            f.seek(0)

            for bline in f:
                line: str = bline.decode('utf-8')

                # Buffer the comments and put them in the output at the end
                if in_comment:
                    # End of the comment, write the buffered contents if they
                    # were not thrown away and if we're not in a disabled
                    # preprocessor branch
                    if blockcomment_end_rx.match(line):
                        assert not multiline_copyright

                        in_comment = False
                        if branch_stack[-1][1] is not False and comment_buffer:
                            out += comment_buffer
                            out += [line]
                            comment_buffer = []

                    # Process the comments further only if we're not in a
                    # disabled preprocessor branch
                    elif branch_stack[-1][1] is not False:
                        # Continuation of a multiline copyright
                        if multiline_copyright:
                            multiline_copyright += line

                            # Final line of a multiline copyright, add it
                            if line.rstrip().endswith('>'):
                                copyrights.add(multiline_copyright)
                                multiline_copyright = None

                        # License block -- extract just copyrights
                        elif copyright_rc.match(line):
                            comment_buffer = []

                            # Add a complete copyright line to the global list
                            if line.rstrip().endswith('>'): copyrights.add(line)

                            # If a multi-line copyright (ending either with a
                            # `,` or a year), wait for next time
                            else: multiline_copyright = line

                        # Otherwise add to the buffer. The buffer is always at least
                        # one line (added when comment start matches), if not then it
                        # means we don't want it (license block, for example)
                        elif comment_buffer: comment_buffer += [line]

                    continue

                # Single-line comment or an empty line
                if linecomment_rx.match(line):
                    # If this is an {{include}} placeholder, finalize the
                    # previous set of includes and open a new one so the new
                    # includes are added to the new placeholder
                    if line.strip() == '// {{includes}}':
                        new_includes += [set()]

                    # Add it only if we're not in a disabled preprocessor
                    # branch and it's either a non-empty line (and comments are
                    # not disabled) or an empty line that's not first in the
                    # file and there's not more than one following each other
                    if branch_stack[-1][1] is not False and ((line.strip() and write_comments) or line.strip() == '// {{includes}}' or (not line.strip() and out and out[-1].strip())):
                        out += [line]
                    continue

                # Start of a multi-line comment
                match = blockcomment_start_rx.match(line)
                if match and '*/' not in line:
                    assert not in_comment
                    assert not comment_buffer

                    in_comment = True

                    # Add the comment only if we're not in a disabled
                    # preprocessor branch and comments are enabled
                    if branch_stack[-1][1] is not False and write_comments:
                        comment_buffer = [line]
                    continue

                # Preprocessor branch
                match = preprocessor_rx.match(line)
                if match:
                    # Unmatched group is None, so use a string instead. Because
                    # we strip the value, we need to put back a space before
                    # the comment.
                    indent = match.group('indent') or ''
                    what = match.group('what')
                    value = match.group('value').strip()
                    comment = match.group('comment') or ''
                    if comment and not comment[0].isspace(): comment = ' ' + comment

                    # Leaf node should always be a real node
                    assert len(branch_stack) >= 1
                    assert branch_stack[-1][0] == True

                    if what in ['if', 'ifdef', 'ifndef', 'elif']:
                        push_value, what, value = simplify_expression(what, value, forced_defines)

                    if what in ['if', 'ifdef', 'ifndef']:
                        # If the parent node disabled visibility, this should
                        # not enable it back
                        if branch_stack[-1][1] is False: push_value = False
                        # Push a new node on the stack
                        branch_stack += [[True, push_value, len(out)]]
                        # If the new node doesn't affect visibility, print it
                        if branch_stack[-1][1] is None:
                            out += ['{}#{} {}{}\n'.format(indent, what, value, comment)]

                    elif what == 'elif':
                        assert len(branch_stack) >= 2
                        # Flip the condition for the else block, if the outer
                        # condition affected visibility, but only if the whole
                        # node isn't already disabled
                        if branch_stack[-1][1] is not None and branch_stack[-2][1] is not False:
                            branch_stack[-1][1] = not branch_stack[-1][1]
                        # If the parent node disabled visibility, this should
                        # not enable it back
                        if branch_stack[-1][1] is False: push_value = False

                        # If the outer branch didn't affect visibility and this
                        # one isn't either, put the processed elif to output.
                        # The outer branch then doesn't need to have its #endif
                        # written, so setting it to True.
                        if branch_stack[-1][1] is None and push_value is None:
                            branch_stack[-1][1] = True # TODO: false? does it matter?
                            out += ['{}#elif {}{}\n'.format(indent, value, comment)]
                        # If the outer branch didn't affect visibility and this
                        # does, put just else to output. That also means we
                        # need to print #endif after, so keep it at None.
                        elif branch_stack[-1][1] is None and push_value is not None:
                            out += ['{}#else{}\n'.format(indent, comment)]
                        # If the outer branch affected visibility and this
                        # does not, put just if (normalized) to output
                        elif branch_stack[-1][1] is not None and push_value is None:
                            out += ['{}#{} {}{}\n'.format(indent, *normalize_expression(value), comment)]
                        # Otherwise not printing anything, meaning both the
                        # outer node and the inner affect visibility
                        else:
                            assert branch_stack[-1][1] is not None and push_value is not None

                        # Turn the parent node into a #elif dummy and add a new
                        # node to handle this new if. We need to do it this way
                        # in order to remember the visibility status of the
                        # else block
                        branch_stack[-1][0] = False
                        branch_stack += [[True, push_value, branch_stack[-1][2]]]
                    elif what == 'else':
                        assert len(branch_stack) >= 2
                        # Put the line to output if the branch didn't affect
                        # visibility. If it did, flip the condition.
                        if branch_stack[-1][1] is None:
                            out += ['{}#else{}\n'.format(indent, comment)]
                        else:
                            branch_stack[-1][1] = not branch_stack[-1][1]
                        # But if the parent node disabled visibility, this
                        # should not enable it back
                        if branch_stack[-2][1] is False:
                            branch_stack[-1][1] = False
                    else:
                        assert what == 'endif'
                        assert len(branch_stack) >= 2
                        # Put the line to output if the branch didn't affect
                        # visibility
                        endif_written = False
                        if branch_stack[-1][1] is None:
                            endif_written = True
                            out += ['{}#endif{}\n'.format(indent, comment)]

                        # Remember the line no. of the `if` statement, so we
                        # can remove the whole thing below if it turns out to
                        # be empty
                        if_lineno = branch_stack[-1][2]
                        branch_stack.pop()

                        # There might be dummy #elif nodes above, drop all of
                        # them until there is a real one. Put an extra endif
                        # in all cases where needed.
                        while branch_stack[-1][0] is False:
                            if branch_stack[-1][1] is None:
                                out += ['{}#endif{}\n'.format(indent, comment)]
                            if_lineno = branch_stack[-1][2]
                            branch_stack.pop()
                        assert len(branch_stack) >= 1

                        # If the endif was written and there's nothing between
                        # the if and the endif, remove the whole thing.
                        if endif_written and if_lineno + 2 == len(out):
                            out.pop()
                            out.pop()

                    continue

                # We're in a preprocessor branch with disabled visibility,
                # ignore anything else that could be inside
                # TODO: still need to handle raw strings, ugh
                if branch_stack[-1][1] is False:
                    continue

                # Include
                match = include_rx.match(line)
                if match:
                    include:str = match.group('file')
                    is_local = match.group('quote') == '"'
                    # Local includes or includes from dependent projects, recurse
                    if (is_local or include.partition('/')[0] in local_include_prefixes) and not include in local_includes_noexpand:
                        # A header corresponding to an implementation file
                        if is_local and '/' not in include:
                            absolute_include = os.path.join(os.path.dirname(file), include)
                            if not os.path.exists(absolute_include): # pragma: no cover
                                for path in paths:
                                    relative_include = os.path.join(path, include)
                                    if os.path.exists(relative_include):
                                        absolute_include = relative_include
                                        break
                            if not os.path.exists(absolute_include):
                                logging.fatal("Can't find %s in %s or any of %s", include, os.path.dirname(file), paths)
                                assert False
                        # Something else
                        else:
                            for path in paths:
                                absolute_include = os.path.join(path, include)
                                if os.path.exists(absolute_include):
                                    break
                            else: # pragma: no cover
                                logging.fatal("Can't find %s in any of %s", include, paths)
                                assert False

                        # If given include file is not parsed yet, parse it
                        if absolute_include not in parsed_files:
                            # If this is the top-level parsed file, add the
                            # contents in-order. Otherwise prepend the include
                            # before this file contents to avoid nested include
                            # guards.
                            parsed_file = parse(absolute_include, level + 1)
                            if toplevel_file == file or tpp_filename.search(absolute_include):
                                out += parsed_file
                            else:
                                includes_out += parsed_file

                    # System or local noexpand include.
                    else:
                        includeline = match.group('include') + '\n'

                        # Already spotted, don't do anything
                        if includeline in all_includes: continue

                        # If the include is wrapped in a preprocessor branch
                        # other than the include guard (i.e., it starts at a
                        # line > 0) and the branch is not disabled/enabled,
                        # then we need to keep it in place -- it might be a
                        # platform-specific thing.
                        if len(branch_stack) > 1 and branch_stack[-1][2] != 0 and branch_stack[-1][1] is None:
                            out += [includeline]

                        # Otherwise add it to the set of not-yet-written
                        # includes, it'll get written to the nearest preceding
                        # {{includes}} placeholder.
                        else:
                            if not new_includes:
                                logging.warning("Includes found before an {{includes}} placeholder, the resulting file will have them on the top")
                                new_includes += [set()]
                            new_includes[-1].add(includeline)

                        # In both cases, add it to the list of global includes
                        # to avoid including it more than once. This is done
                        # with the assumption that platform-specific includes
                        # always get wrapped in the preprocessor branch the
                        # same way.
                        # TODO: this might cause problems?
                        all_includes.add(includeline)
                    continue

                # Pragma
                match = acme_pragma_rx.match(line)
                if match:
                    what = match.group('what')
                    value = match.group('value').rstrip()
                    if what == 'enable':
                        forced_defines[value] = True
                    elif what == 'disable':
                        forced_defines[value] = False
                    elif what == 'path':
                        paths += [os.path.join(base_directory, value)]
                    elif what == 'local':
                        local_include_prefixes += [value]
                    elif what == 'comments':
                        write_comments = value == 'on'
                    elif what == 'revision':
                        path, _, command = value.partition(' ')
                        revision_commands[path] = command.strip()
                    elif what == 'stats':
                        id, _, command = value.partition(' ')
                        stats_commands[id] = command.strip()
                    elif what == 'noexpand':
                        local_includes_noexpand.add(value)
                    else:
                        logging.warning("Unknown #pragma ACME %s %s", what, value)

                    continue

                # Preprocessor define -- if it's among the forced ones, ignore
                match = define_rx.match(line)
                if match and match.group('name') in forced_defines:
                    continue

                # Something else, copy verbatim to the output. Strip the
                # trailing comment, if requested
                if line.rstrip().endswith('*/'):
                    out += [line[:line.rindex('/*')].rstrip() + '\n']
                else:
                    out += [line]

        # Assert that the branch stack is correct
        assert len(branch_stack) == 1

        # Drop empty liness off the end
        while out and not out[-1].strip(): out.pop()

        # Return parsed lines, stuff from includes in front
        return includes_out + out

    lines = parse(toplevel_file, 0)

    # For each include placeholder put the corresponding includes there. If
    # there's none, put them on the top.
    if new_includes:
        i = 0
        while i != len(lines):
            if lines[i].strip() == '// {{includes}}':
                sorted_includes = sort_includes(new_includes[0])
                lines = lines[:i] + sorted_includes + lines[i + 1:]
                new_includes.pop(0)
                if not new_includes: break
                i = i + len(sorted_includes)
            else:
                i = i + 1
        else:
            # Warning already printed when new_includes was discovered to be empty
            lines = sort_includes(new_includes[0]) + lines

    # Find a copyright placeholder and put the copyrights there
    if copyrights:
        for i, line in enumerate(lines):
            if line.strip() == '{{copyright}}':
                lines = lines[:i] + sort_copyrights(copyrights) + lines[i + 1:]
                break
        else:
            logging.warning(" No {{copyrights}} placeholder found, ignoring found copyright statements")

    # If no custom revision fetch command was provided, add a default one. Then
    # for each revision path find a corresponding placeholder and replace it.
    if '*' not in revision_commands:
        revision_commands['*'] = 'git describe --dirty --always'
    for path, command in revision_commands.items():
        placeholder = '{{{{revision{}}}}}'.format('' if path == '*' else ':' + path)
        revision = None
        for i, line in enumerate(lines):
            if not placeholder in line: continue
            if not revision:
                # Find the file where the revision should be fetched
                if path == '*': cwd = os.path.dirname(os.path.realpath(toplevel_file))
                else:
                    for file in parsed_files:
                        realfile = os.path.realpath(file)
                        if path in realfile:
                            cwd = os.path.dirname(realfile)
                            break
                    else: # pragma: no cover
                        logging.fatal("No matching file found for expanding %s", placeholder)
                        assert False

                revision = subprocess.check_output(command, cwd=cwd, shell=True).decode('utf-8').strip()
            lines[i] = line.replace(placeholder, revision)

    # Create the output directory
    output_dir = os.path.dirname(output)
    if not os.path.exists(output_dir): os.makedirs(output_dir)

    # Perform some stats on file contents, passing them to stdin, running in
    # the (freshly created) output directory
    for id, command in stats_commands.items():
        placeholder = '{{{{stats:{}}}}}'.format(id)
        stats = None
        for i, line in enumerate(lines):
            if not placeholder in line: continue
            if not stats:
                stats = subprocess.check_output(command, cwd=output_dir, input=''.join(lines).encode('utf-8'), shell=True).decode('utf-8').strip()
            lines[i] = line.replace(placeholder, stats)

    logging.info('Writing %i lines to %s', len(lines), output)
    with open(output, 'wb') as of:
        for line in lines:
            of.write(line.encode('utf-8'))

if __name__ == '__main__': # pragma: no cover
    parser = argparse.ArgumentParser(
        description=r"""Creates single-header libraries from given top-level input file.""",
        epilog="""If output exists and is a directory, the file is saved inside with the same name as the input.""")
    parser.add_argument('file', help='top-level file')
    parser.add_argument('-o', '--output', help="output file or directory", default='output')
    parser.add_argument('--debug', help="verbose debug output", action='store_true')
    args = parser.parse_args()

    if args.debug:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)

    output = os.path.join(os.path.dirname(args.file), args.output)

    # If a directory, save the file inside
    if os.path.exists(output) and os.path.isdir(output):
        output = os.path.join(output, os.path.basename(args.file))

    # Otherwise assume it's a filename, create all directories above it if they
    # don't exist
    else: os.makedirs(os.path.dirname(output), exist_ok=True)

    acme(args.file, output)



#!/usr/bin/env python3
import unittest
import numpy as np

def fetchLoad(loader, fetchfile, **kwds):
    """Fetch remote binary file from the :py:mod:`brille` repository

    Parameters
    ----------
    loader : function
        The function which loads the requested file into Python
    fetchfile : str
        The filename (with extension) of the file to fecth
    **kwds :
        All keyword arguments are passed to the `loader` function

    Returns
    -------
    loaded
        The loaded object or value returned by the `loader` function
    """
    import tempfile
    import requests
    import pathlib
    base_url = "https://raw.githubusercontent.com/brille/brille/master/wrap/tests"
    with tempfile.TemporaryDirectory() as tmp_dir:
        r = requests.get(base_url + "/" + fetchfile)
        if not r.ok:
            raise Exception("Fetching {} failed with reason '{}'".format(fetchfile, r.reason))
        out_path = str(pathlib.Path(tmp_dir, fetchfile))
        open(out_path, 'wb').write(r.content)
        return loader(out_path, **kwds)


def getLoad(loader, file, **kwds):
    """Load a binary file

    If the file is not located in the `brille` module directory this function
    will attempt to obtain it from the `brille` repository over any available
    network connection.

    Parameters
    ----------
    loader : function
        The function which loads the requested file into Python
    fetchfile : str
        The filename (with extension) of the file to fecth
    **kwds :
        All keyword arguments are passed to the `loader` function

    Returns
    -------
    loaded
        The loaded object or value returned by the `loader` function
    """
    import os
    import pathlib
    docs_dir = os.path.dirname(os.path.abspath(__file__))
    try:
        return loader(str(pathlib.Path(docs_dir, file)), **kwds)
    except FileNotFoundError:
        print('{} not found in {}. Fetching remote content.'.format(file, docs_dir))
        return fetchLoad(loader, file, **kwds)


class GammaTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        # Set up nacl test data, lattice and grid

        from brille import Basis, Symmetry, Lattice, BrillouinZone, BZTrellisQdc
        # load numpy arrays from the compressed binary pack
        nacl = getLoad(np.load, 'test_5_gamma.npz')
        # construct the NaCl direct lattice
        bas = Basis(nacl['atom_positions'], nacl['atom_index'])
        sym = Symmetry(nacl['spacegroup_mat'], nacl['spacegroup_vec'])
        lat = Lattice(nacl['basis_vectors'], symmetry=sym, basis=bas)

        # use it to construct an irreducible Brillouin zone
        bz = BrillouinZone(lat)
        # and use that to produce a hybrid interpolation grid
        # with parameters stored in the binary pack
        max_volume = float(nacl['grid_max_volume'])
        always_triangulate = bool(nacl['grid_always_triangulate'])
        grid = BZTrellisQdc(bz, max_volume, always_triangulate)

        cls.nacl = nacl
        cls.lat = lat
        cls.grid = grid

    def test_nacl(self):
        # Verify interpolation with defined test data/lattice/grid
        nacl = self.nacl
        lat = self.lat
        grid = self.grid

        # verify the stored grid points to ensure we have the same irreducible
        # wedge and grid
        # Though the points could be permuted:
        perm = np.hstack([np.argwhere(np.all(np.isclose(nacl['grid_rlu'], x), axis=1)) for x in grid.rlu]).flatten()
        self.assertTrue(np.allclose(nacl['grid_rlu'][perm], grid.rlu))

        # insert the Euphonic-derived eigenvalues and eigenvectors for the grid
        # points, which are used in the interpolation
        # Original vectors elements enumeration was [0, 24 0, 3, 0, 0] where
        # RotatesLike::Gamma = 3. New interface includes extra enumeration
        # LengthUnit::real_lattice = 3, and the RotatesLike enumeration has
        # changed, so RotatesLike::Gamma = 2 now. Replace here rather than
        # regenerate file.
        vec_els = np.array([0, 24, 0, 2, 3, 0, 0])
        grid.fill(
            nacl['grid_values'][perm], nacl['grid_values_elements'], nacl['grid_values_weights'],
            nacl['grid_vectors'][perm], vec_els, nacl['grid_vectors_weights'],
            bool(nacl['grid_sort']))

        # the fourth grid point is inside of the irreducible volume, which
        # ensures we avoid degeneracies
        q_ir = nacl['grid_rlu'][4:5]
        # find all symmetry equivalent q within the first Brillouin zone by
        # applying each pointgroup operator to q_ir to find q_nu = R^T_nu q_ir
        q_nu = np.einsum('xji,aj->xi', lat.pointgroup.W, q_ir)

        # The std::sort algorithm does not provide the same pointgroup sorting
        # on all systems, but there should be a permutation mapping:
        perm = np.array([np.squeeze(np.argwhere(np.all(np.isclose(q_nu, x), axis=1))) for x in nacl['q_nu']])
        # The permutation must be complete and unique
        self.assertTrue(np.unique(perm).size == q_nu.shape[0])
        # And the permuted q_nu must match the stored q_nu
        q_nu = q_nu[perm]
        self.assertTrue(np.allclose(q_nu, nacl['q_nu']))

        # Use the grid to interpolate at each q_nu:
        br_val, br_vec = grid.ir_interpolate_at(q_nu)
        br_val = np.squeeze(br_val)

        # verify that q_ir does not have degeneracies:
        self.assertFalse(np.any(np.isclose(np.diff(br_val, axis=1), 0.)))
        # and that the 'interpolated' eigenvalues are identical for all q_nu
        self.assertTrue(np.allclose(np.diff(br_val, axis=0), 0.))
        # plus that the interpolated eigenvalues match the store Euphonic eigenvalues
        self.assertTrue(np.allclose(br_val, nacl['euphonic_values']))

        # convert the eigenvectors into the same cartesian coordinate system
        # used by Euphonic
        br_vec = np.einsum('ba,ijkb->ijka', nacl['basis_vectors'], br_vec)
        # load the Euphonic calculated eigenvectors
        eu_vec = nacl['euphonic_vectors']
        # The 'interpolated' eigenvectors and the Euphonic eigenvectors should
        # only be equivalent up to an overall phase factor, so find it:
        antiphase = np.exp(-1J * np.angle(np.einsum('qmij,qmij->qm', np.conj(eu_vec), br_vec)))
        # and remove the phase from the interpolated eigenvectors
        br_vec = np.einsum('ab,abij->abij', antiphase, br_vec)
        # now all eigenvectors must match
        self.assertTrue(np.allclose(br_vec, eu_vec))

        # as an extra check we *could* verfiy that the output of brille is
        # unchanged, but we don't care about changes in overall phase factor

    def test_cartesian_eigenvector_interpolation(self):
        # Verify interpolation with Cartesian eigenvectors
        nacl = self.nacl
        grid = self.grid

        # Permuted grid points:
        perm = np.hstack([
            np.argwhere(np.all(np.isclose(nacl['grid_rlu'], x), axis=1))
                for x in grid.rlu]
            ).flatten()

        # Convert input grid vectors from basis to Cartesian
        grid_vecs_cart =  np.einsum(
            'ba,ijkb->ijka', nacl['basis_vectors'], nacl['grid_vectors'])
        # Use RotatesLike::Gamma = 2 and LengthUnit::angstrom = 1
        vec_els = np.array([0, 24, 0, 2, 1, 0, 0])
        grid.fill(
            nacl['grid_values'][perm], nacl['grid_values_elements'],
            nacl['grid_values_weights'], grid_vecs_cart[perm],
            vec_els, nacl['grid_vectors_weights'])

        # Use the grid to interpolate at each q_nu:
        br_val, br_vec = grid.ir_interpolate_at(nacl['q_nu'])

        # load the Euphonic calculated eigenvectors
        eu_vec = nacl['euphonic_vectors']
        # The 'interpolated' eigenvectors and the Euphonic eigenvectors should
        # only be equivalent up to an overall phase factor, so find it:
        antiphase = np.exp(-1J * np.angle(
            np.einsum('qmij,qmij->qm', np.conj(eu_vec), br_vec)))
        # and remove the phase from the interpolated eigenvectors
        br_vec = np.einsum('ab,abij->abij', antiphase, br_vec)
        # now all eigenvectors must match
        self.assertTrue(np.allclose(br_vec, eu_vec))

    def test_unsupported_rotateslike_lengthunit_combinations_error(self):
        nacl = self.nacl
        grid = self.grid
        # RotatesLike, Lengthunit
        combos = [(0, 1), (1, 0), (2, 2), (2, 4)]

        for combo in combos:
            vec_els = np.array([0, 24, 0, *combo, 0, 0])
            grid.fill(nacl['grid_values'], nacl['grid_values_elements'],
            nacl['grid_values_weights'], nacl['grid_vectors'],
            vec_els, nacl['grid_vectors_weights'])
            with self.assertRaises(RuntimeError):
                grid.ir_interpolate_at(nacl['q_nu'])


if __name__ == '__main__':
    unittest.main()

#
# # The data used in  the above test was generated using NDLT1145, a Dell XPS
# # laptop running Windows 10, using
# #   brille   v0.5.0-master.1b5e5a6
# #   brilleu  v0.2.3
# #   euphonic v0.3.0
# # and the following script:
#
# import brille as b, numpy as np, euphonic, brilleu
# from brilleu.brilleu import getBrillEuObj
#
# nacl = {}
# nacl['brille_version'] = b.version
# nacl['brilleu_version'] = brilleu.__version__
# nacl['euphonic_version'] = euphonic.__version__
# nacl['grid_sort'] = False
# nacl['grid_max_volume'] = 0.1
#
# breu = getBrillEuObj('NaCl', sort=nacl['grid_sort'], max_volume=nacl['grid_max_volume'], parallel=False)
#
# nacl['basis_vectors'] = breu.grid.BrillouinZone.lattice.star.lattice_matrix
# nacl['atom_positions'] = breu.crystal.atom_positions
# nacl['atom_index'] = breu.crystal.atom_index
# nacl['spacegroup_mat'] = breu.crystal.symmetry.W
# nacl['spacegroup_vec'] = breu.crystal.symmetry.w
#
# nacl['grid_always_triangulate'] = False
# nacl['grid_rlu'] = breu.grid.rlu
# nacl['grid_values'] = breu.grid.values
# nacl['grid_values_elements'] = (1.,)
# nacl['grid_values_weights'] = (13605.693, 0., 0.)
# nacl['grid_vectors'] = breu.grid.vectors
# nacl['grid_vectors_elements'] = (0, 3*breu.data.crystal.n_atoms, 0, 3, 0, 0)
# nacl['grid_vectors_weights'] = (0., 1., 0.)
#
# q_ir = breu.grid.rlu[4:5]
# nacl['pointgroup'] = breu.grid.BrillouinZone.lattice.pointgroup.W
# nacl['q_nu'] = np.einsum('xji,aj->xi', nacl['pointgroup'], q_ir)
#
# br_nu = breu.QpointPhononModes(nacl['q_nu'], interpolate=True)
# nacl['brille_values'] = br_nu.ω
# nacl['brille_vectors'] = br_nu.ε
#
# eu_nu = breu.QpointPhononModes(nacl['q_nu'], interpolate=False)
# nacl['euphonic_values'] = eu_nu.ω
# nacl['euphonic_vectors'] = eu_nu.ε
#
# np.savez('test_5_gamma.npz', **nacl)
